// Built with Microsoft Visual C++ 2008 Express
// Configuration: Project->Properties->Configuration Properties->
//				Linker->Input->Additional Dependencies = Vfw32.lib

#include <windows.h>
#include <vfw.h>
#include <math.h>		// for sqrt
#include <stdio.h>		// for swprintf

#define IDT_CAPTURE 5000
#define IDT_SERIAL 5001

int COLOR_CUT = 60;						// tolerance for color distance, higher is more permissive
const float DARK_CUT = (float)0.15;		// darkness that is considered black, lower is more permissive	
const int NOISE_CUT = 4;				// number of neighbors with which a pixel needs to survive
const int NOISE_REPEAT = 1;				// number of times to repeat the noise removal routine
const float EYE_SPACING = (float)0.14;	// distance between cameras based on original packaging (in meters)

HANDLE hComm = NULL;
HWND hwndMain = NULL, hwndCam1 = NULL, hwndCam2 = NULL;

BYTE *FRAME[2] = {NULL, NULL};
BYTE *SNAPSHOT[2] = {NULL, NULL};
POINT CANVAS[2] = {{0, 0}, {0, 0}};
POINT CENTER[2] = {{0, 0}, {0, 0}};
POINT AVG_CENTER[2] = {{0, 0}, {0, 0}};
int DISPLAY_MODE = 0;
int X1_SUM = 0, Y1_SUM = 0, X2_SUM = 0, Y2_SUM = 0, DIST_COUNT = 0;

void RemoveNoise(int index, int iterations)
{
	int i, j, k, c;

	if(FRAME[index] == NULL)
		return;

	for(k = 0; k < iterations; ++k)
		for(i = 0; i < CANVAS[index].y; ++i)
			for(j = 0; j < CANVAS[index].x; ++j)
			{
				c = 0;
				if(i == 0 || j ==0 || i == CANVAS[index].y-1 || j == CANVAS[index].x-1)
					continue;		// ignore border pixels
				if(FRAME[index][3*((i+1)*CANVAS[index].x+j)] == 255) ++c;
				if(FRAME[index][3*((i+1)*CANVAS[index].x+j+1)] == 255) ++c;
				if(FRAME[index][3*(i*CANVAS[index].x+j+1)] == 255) ++c;
				if(FRAME[index][3*((i-1)*CANVAS[index].x+j+1)] == 255) ++c;
				if(FRAME[index][3*((i-1)*CANVAS[index].x+j)] == 255) ++c;
				if(FRAME[index][3*((i-1)*CANVAS[index].x+(j-1))] == 255) ++c;
				if(FRAME[index][3*(i*CANVAS[index].x+(j-1))] == 255) ++c;
				if(FRAME[index][3*((i+1)*CANVAS[index].x+j-1)] == 255) ++c;
				if(c > 8-NOISE_CUT)
				{
					FRAME[index][3*(i*CANVAS[index].x+j)] = 255;
					FRAME[index][3*(i*CANVAS[index].x+j)+1] = 255;
					FRAME[index][3*(i*CANVAS[index].x+j)+2] = 255;
				}
			}
}

// r, g, b are between 0 and 1
// brightness is defined to be the length of the vector
// we normalize the length to one
void NormalizeBrightness(float *r, float *g, float *b)
{
	float len = (float)sqrt((*r)*(*r)+(*g)*(*g)+(*b)*(*b));
	if(len > DARK_CUT)	// only if it is not black
	{
		*r = *r/len;
		*g = *g/len;
		*b = *b/len;
	}
}

// converts buffered frames to grayscale based on how far each
// pixel's color is from the specified color
// note: need to normalize to a certain brightness
// problem: normalizing brightness means black could be seen as red
void ColorDistance(int index, COLORREF color)
{
	int i, j;
	float dist, r, g, b;
	float R = (float)(GetRValue(color))/255;
	float G = (float)(GetGValue(color))/255;
	float B = (float)(GetBValue(color))/255;

	if(FRAME[index] == NULL)
		return;

	NormalizeBrightness(&R, &G, &B);

	for(i = 0; i < CANVAS[index].y; ++i)
		for(j = 0; j < CANVAS[index].x; ++j)
		{
			r = (float)(FRAME[index][3*(i*CANVAS[index].x+j)])/255;
			g = (float)(FRAME[index][3*(i*CANVAS[index].x+j)+1])/255;
			b = (float)(FRAME[index][3*(i*CANVAS[index].x+j)+2])/255;
			NormalizeBrightness(&r, &g, &b);
			dist = 255*sqrt((float)((r-R)*(r-R)+(g-G)*(g-G)+(b-B)*(b-B)))/sqrt((float)5.0);
			if(dist > COLOR_CUT)
				dist = 255;
			FRAME[index][3*(i*CANVAS[index].x+j)] = (int)dist;
			FRAME[index][3*(i*CANVAS[index].x+j)+1] = (int)dist;
			FRAME[index][3*(i*CANVAS[index].x+j)+2] = (int)dist;
		}
}

void InitSerial()
{
    wchar_t msg[64];
    DCB serial;
    hComm = CreateFile(L"COM1", GENERIC_READ | GENERIC_WRITE, 0, 0, 
                    OPEN_EXISTING, 0/*FILE_FLAG_OVERLAPPED*/, 0);
    if (hComm == INVALID_HANDLE_VALUE || hComm == NULL)
    {
        hComm = NULL;
        MessageBox(hwndMain, L"Could not open serial port!", L"Parallax", MB_OK);
        return;
    }
    serial.DCBlength = sizeof(DCB);
    if(GetCommState(hComm, &serial) == 0)
    {
        MessageBox(hwndMain, L"Could not read serial port parameters!", L"Parallax", MB_OK);
        return;
    }
    serial.BaudRate = CBR_9600;
    serial.ByteSize = 8;
    serial.Parity = NOPARITY;
    serial.StopBits = ONESTOPBIT;
    if(SetCommState(hComm, &serial) == 0)
    {
        swprintf(msg, sizeof(msg)/sizeof(wchar_t), 
            L"Could not set serial port parameters (Error %d)!", GetLastError());
        MessageBox(hwndMain, msg, L"Parallax", MB_OK);
    }
}

// Send coordinates of the centers of mass over the serial port
void SendSerial(int x1, int y1, int x2, int y2)
{
    DWORD sent = 0;
    char str[64];

    if(hComm == NULL)
        return;

    sprintf(str, " %d %d %d %d", x1, y1, x2, y2);
    WriteFile(hComm, str, (DWORD)strlen(str), &sent, NULL);
}

float CalculateDistance()
{
	float pixelWidth = (float)(CANVAS[0].x+CANVAS[1].x)/2;
	float delta = abs((float)(AVG_CENTER[0].x - AVG_CENTER[1].x));

    if(AVG_CENTER[0].x == 0 && AVG_CENTER[0].y == 0)
        return -1.0;
    if(AVG_CENTER[1].x == 0 && AVG_CENTER[1].y == 0)
        return -1.0;
	if(AVG_CENTER[0].x - AVG_CENTER[1].x == 0)
		return -1.0;		// object is too far away to tell where it is
	return (float)EYE_SPACING*pixelWidth/(2*delta*tan((float)0.524/2));
}

POINT CenterOfMass(int index)
{
	POINT ret = {0, 0};
	int i, j;
	int xsum = 0, ysum = 0, count = 0;

	if(FRAME[index] == NULL)
		return ret;

	for(i = 0; i < CANVAS[index].y; ++i)
		for(j = 0; j < CANVAS[index].x; ++j)
			if(FRAME[index][3*(i*CANVAS[index].x+j)] < 255)
			{
				xsum += j;
				ysum += i;
				count += 1;
			}
	if(count == 0)
		return ret;
	ret.x = xsum/count;
	ret.y = ysum/count;
	return ret;
}

LRESULT CALLBACK GrabFrame(HWND hWnd, LPVIDEOHDR lpVHdr, int index)
{
	unsigned int i;
	DWORD length = lpVHdr->dwBytesUsed;
	LPBYTE camframe = lpVHdr->lpData;

	if(FRAME[index] == NULL)
		FRAME[index] = (BYTE*)malloc(3*CANVAS[index].x*CANVAS[index].y*sizeof(BYTE));
	if(SNAPSHOT[index] == NULL)
		SNAPSHOT[index] = (BYTE*)malloc(3*CANVAS[index].x*CANVAS[index].y*sizeof(BYTE));
	for(i = 0; i < length && i < 3*CANVAS[index].x*CANVAS[index].y; ++i)
	{
		FRAME[index][i] = camframe[i];
		SNAPSHOT[index][i] = camframe[i];
	}
	return TRUE;
} 

LRESULT CALLBACK GrabFrame1(HWND hWnd, LPVIDEOHDR lpVHdr)
{
	return GrabFrame(hWnd, lpVHdr, 0);
}

LRESULT CALLBACK GrabFrame2(HWND hWnd, LPVIDEOHDR lpVHdr)
{
	return GrabFrame(hWnd, lpVHdr, 1);
}

HWND InitCamera(HWND hwndParent, int xpos, int ypos, int driver)
{
	static int count = 0;
	HWND hwndRet;
	CAPSTATUS CapStatus = {0};

	hwndRet = capCreateCaptureWindow(L"Camera", WS_CHILD, 
			xpos, ypos, 320, 240, hwndParent, ++count);
	if(hwndRet == NULL)
		return NULL;
	capDriverConnect(hwndRet, driver);
	capGetStatus(hwndRet, &CapStatus, sizeof(CAPSTATUS));
	CANVAS[count-1].x = CapStatus.uiImageWidth;
	CANVAS[count-1].y = CapStatus.uiImageHeight;
	return hwndRet;
}

void DrawFrame(HDC hdc, int x, int y, int index, int mode)
{
    int i, j;
	BITMAPINFO bmi = {0};
    HBRUSH brush;
    RECT rc;

	bmi.bmiHeader.biSize = sizeof(bmi.bmiHeader);
	bmi.bmiHeader.biPlanes = 1;
	bmi.bmiHeader.biBitCount = 24;
	bmi.bmiHeader.biWidth = CANVAS[index].x;
	bmi.bmiHeader.biHeight = CANVAS[index].y;
	bmi.bmiHeader.biSizeImage = 3*CANVAS[index].x*CANVAS[index].y;
	
	brush = CreateSolidBrush(RGB(0, 255, 0));
    if(mode == 0 && SNAPSHOT[index] != NULL)
        SetDIBitsToDevice(hdc, x, y, CANVAS[index].x, CANVAS[index].y, 
		    0, 0, 0, CANVAS[index].y, SNAPSHOT[index], &bmi, DIB_RGB_COLORS);
    if(mode == 1 && FRAME[index] != NULL)
        SetDIBitsToDevice(hdc, x, y, CANVAS[index].x, CANVAS[index].y, 
		    0, 0, 0, CANVAS[index].y, FRAME[index], &bmi, DIB_RGB_COLORS);
    if(mode == 2 && SNAPSHOT[index] != NULL && FRAME[index] != NULL)
    {
        SetDIBitsToDevice(hdc, x, y, CANVAS[index].x, CANVAS[index].y, 
		    0, 0, 0, CANVAS[index].y, SNAPSHOT[index], &bmi, DIB_RGB_COLORS);

	    for(i = 0; i < CANVAS[index].y; ++i)
		    for(j = 0; j < CANVAS[index].x; ++j)
			    if(FRAME[index][3*(i*CANVAS[index].x+j)] < 255)
                    SetPixel(hdc, x+j, y+(CANVAS[index].y-i), RGB(255, 255, 255));
    }
    
    rc.top = y + CANVAS[index].y - CENTER[index].y - 3;
	rc.bottom = y + CANVAS[index].y - CENTER[index].y + 3;
	rc.left = x + CENTER[index].x - 3;
	rc.right = x + CENTER[index].x + 3;
	FillRect(hdc, &rc, brush);
	DeleteObject(brush);
}


LRESULT CALLBACK MainWndProc(HWND hWnd, UINT msg, WPARAM wParam, LPARAM lParam)
{
	HDC hdc;
	PAINTSTRUCT ps;
    RECT rc;

	float distance;
	wchar_t out[64];
	wchar_t err[] = L"Distance: Too Far!";

	switch(msg)
	{
	case WM_DESTROY:
		PostQuitMessage(0);
		break;
	case WM_TIMER:
		if(wParam == IDT_CAPTURE)
		{
			capSetCallbackOnFrame(hwndCam1, GrabFrame1);
			capGrabFrameNoStop(hwndCam1);
			ColorDistance(0, RGB(0, 0, 255));
			RemoveNoise(0, NOISE_REPEAT);
			CENTER[0] = CenterOfMass(0);
			capSetCallbackOnFrame(hwndCam2, GrabFrame2);
			capGrabFrameNoStop(hwndCam2);
			ColorDistance(1, RGB(0, 0, 255));
			RemoveNoise(1, NOISE_REPEAT);
			CENTER[1] = CenterOfMass(1);
            X1_SUM += CENTER[0].x;
            Y1_SUM += CENTER[0].y;
            X2_SUM += CENTER[1].x;
            Y2_SUM += CENTER[1].y;
            DIST_COUNT += 1;
			InvalidateRect(hWnd, NULL, 0);
			UpdateWindow(hWnd);
		}
        if(wParam == IDT_SERIAL)
        {
            if(DIST_COUNT > 0)
            {
                SendSerial(X1_SUM/DIST_COUNT, Y1_SUM/DIST_COUNT, 
                    X2_SUM/DIST_COUNT, Y2_SUM/DIST_COUNT);
                AVG_CENTER[0].x = X1_SUM/DIST_COUNT;
                AVG_CENTER[0].y = Y1_SUM/DIST_COUNT;
                AVG_CENTER[1].x = X2_SUM/DIST_COUNT;
                AVG_CENTER[1].y = Y2_SUM/DIST_COUNT;
                X1_SUM = Y1_SUM = X2_SUM = Y2_SUM = DIST_COUNT = 0;
            }
        }
		break;
    case WM_KEYDOWN:
        if(wParam == VK_UP)
            COLOR_CUT++;
        if(wParam == VK_DOWN)
            COLOR_CUT--;
        break;
	case WM_LBUTTONDOWN:
        DISPLAY_MODE = (DISPLAY_MODE+1)%3;
		break;
	case WM_PAINT:
		hdc = BeginPaint(hWnd, &ps);

        DrawFrame(hdc, 0, 0, 0, DISPLAY_MODE);
        DrawFrame(hdc, CANVAS[0].x, 0, 1, DISPLAY_MODE);

		rc.top = 10;
		rc.bottom = 40;
		rc.left = 10;
		rc.right = 150;
		distance = CalculateDistance();

		if(distance == -1.0)
			DrawText(hdc, err, (int)wcslen(err), &rc, DT_CENTER);
		else
		{
			swprintf(out, sizeof(out)/sizeof(wchar_t), L"Distance: %f", distance);
			DrawText(hdc, out, (int)wcslen(out), &rc, DT_CENTER);
		}
		EndPaint(hWnd, &ps);
		break;
	default:
		return DefWindowProc(hWnd, msg, wParam, lParam);
	}
	return FALSE;
}

void ClientResize(HWND hWnd, int width, int height)
{
	RECT rcClient, rcWindow;
	POINT ptDiff;

	GetClientRect(hWnd, &rcClient);
	GetWindowRect(hWnd, &rcWindow);
	ptDiff.x = (rcWindow.right - rcWindow.left) - rcClient.right;
	ptDiff.y = (rcWindow.bottom - rcWindow.top) - rcClient.bottom;
	MoveWindow(hWnd,rcWindow.left, rcWindow.top, width + ptDiff.x, height + ptDiff.y, TRUE);
}

int APIENTRY WinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance, LPSTR lpCmdLine, int nCmdShow)
{
	MSG msg;
	WNDCLASSEX wndclass = {0};

	wndclass.cbSize = sizeof(wndclass);
	wndclass.lpszClassName = L"Parallax";
	wndclass.hInstance = hInstance;
	wndclass.lpfnWndProc = MainWndProc;
	wndclass.hCursor = LoadCursor(NULL, IDC_ARROW);
	wndclass.hbrBackground = (HBRUSH)(COLOR_BACKGROUND);


	if(!RegisterClassEx(&wndclass))
	{
		MessageBox(NULL, L"Could not register window class!", L"Error", MB_ICONEXCLAMATION);
		return 0;
	}

	hwndMain = CreateWindowEx(WS_EX_CLIENTEDGE, L"Parallax", L"Parallax", 
		WS_OVERLAPPEDWINDOW | WS_VISIBLE, CW_USEDEFAULT, CW_USEDEFAULT, 640, 240, NULL, NULL, hInstance, NULL);

	if(hwndMain == NULL)
	{
		MessageBox(NULL, L"Could not create main window!", L"Error", MB_ICONEXCLAMATION);
		return 0;
	}

	hwndCam1 = InitCamera(hwndMain, 0, 0, 0);
	hwndCam2 = InitCamera(hwndMain, 320, 0, 0);

	ClientResize(hwndMain, CANVAS[0].x + CANVAS[1].x, max(CANVAS[0].y, CANVAS[1].y));

	if(!hwndCam1 || !hwndCam2)
		MessageBox(hwndMain, L"Could not initialize capture windows!", L"Error", MB_ICONEXCLAMATION);

    InitSerial();

	SetTimer(hwndMain, IDT_CAPTURE, 40, (TIMERPROC)NULL);
    SetTimer(hwndMain, IDT_SERIAL, 1000, (TIMERPROC)NULL);

	while(GetMessage(&msg, NULL, 0, 0) != 0) 
	{
		TranslateMessage(&msg);
		DispatchMessage(&msg);
	}

	return 0;
}
